from pdlpy.distribution import Distribution
from pdlpy.math import e, erf, pi, sqrt


class Normal(Distribution):
    """
    Continuous probability distribution of the random variable X that is assumed to be additively produced by many small effects
    """

    def __init__(self, mean: float, var: float):
        """
        Paramters
        mean: the expectation of the distribution
        var: the variance of the distribution
        """
        self._mean = mean
        self._median = mean
        self._mode = mean
        self._var = var

    def __str__(self):
        return (
            "Normal("
            f"mean={self._mean:.2f}, "
            f"median={self._median:.2f}, "
            f"mode={self._mode:.2f}, "
            f"var={self._var:.2f}"
            ")"
        )

    def pdf(self, x: float) -> float:
        """
        Probability Density Function

        Paramters
        x: a value of random variable X

        Returns
        the relative likelihood that a value of X would lie in sample space
        """
        return (1 / sqrt(2 * pi * self._var)) * e ** (
            -((x - self._mean) ** 2 / 2 * self._var)
        )

    def cdf(self, x: float) -> float:
        """
        Cumulative Distribution Function

        Parameters
        x: a value of the random variable X

        Returns
        the probability that X will take a value less than or equal to x
        """
        return (1 + erf((x - self._mean) / (sqrt(self._var * 2)))) / 2



from pdlpy.distribution import Distribution
from pdlpy.math import ceil, log2


class Geometric(Distribution):
    """
    Discrete probability distributions of the random number X of Bernoulli trials needed to get a single success
    """

    def __init__(self, p: float):
        """
        Parameters
        p: the probability of the positive outcome of the experiment
        """
        self._p = p
        self._mean = 1 / p
        self._median = ceil(-1 / log2(1 - p)) if p != 1 else 1
        self._mode = 1
        self._var = (1 - p) / (p ** 2)

    def __str__(self):
        return (
            "Geometric("
            f"p={self._p:.2f}, "
            f"mean={self._mean:.2f}, "
            f"median={self._median:.2f}, "
            f"mode={self._mode:.2f}, "
            f"var={self._var:.2f}"
            ")"
        )

    def pmf(self, x: int) -> float:
        """
        Probability Mass Function

        Parameters
        x: a value of the random variable X

        Returns
        the probability that X will take a value exactly equal to x
        """
        return (1 - self._p) ** x * self._p

    def cdf(self, x: int) -> float:
        """
        Cumulative Distribution Function

        Parameters
        x: a value of the random variable X

        Returns
        the probability that X will take a value less than or equal to x
        """
        if x == 0:
            return self.pmf(0)
        else:
            return self.pmf(x) + self.cdf(x - 1)



from pdlpy.distribution import Distribution
from pdlpy.math import e, log


class Exponential(Distribution):
    """
    Continuous probability distribution of time between events in a Poisson process
    """

    def __init__(self, rate: float):
        """
        Paramters
        rate: the average number of events
        """
        self._rate = rate
        self._mean = rate ** -1
        self._median = log(2) / rate
        self._mode = 0
        self._var = rate ** -2

    def __str__(self):
        return (
            "Exponential("
            f"rate={self._rate:.2f}, "
            f"mean={self._mean:.2f}, "
            f"median={self._median:.2f}, "
            f"mode={self._mode:.2f}, "
            f"var={self._var:.2f}"
            ")"
        )

    def pdf(self, x: float) -> float:
        """
        Probability Density Function

        Paramters
        x: a value of random variable X

        Returns
        the relative likelihood that a value of X would lie in sample space
        """
        return self._rate * e ** (-self._rate * x)

    def cdf(self, x: float) -> float:
        """
        Cumulative Distribution Function

        Parameters
        x: a value of the random variable X

        Returns
        the probability that X will take a value less than or equal to x
        """
        return 1 - e ** (-self._rate * x)



from pdlpy.distribution import Distribution


class Bernoulli(Distribution):
    """
    Discrete probability distribution of a random variable X which takes either value 1 or 0
    """

    def __init__(self, p: float):
        """
        Parameters
        p: the probability of positive outcome of an experiment
        """
        self._p = p
        self._mean = p
        self._median = 0 if p < 0.5 else 1
        self._mode = 0 if p < 0.5 else 1
        self._var = p * (1 - p)

    def __str__(self):
        return (
            "Bernoulli("
            f"p={self._p:.2f}, "
            f"mean={self._mean:.2f}, "
            f"median={self._median:.2f}, "
            f"mode={self._mode:.2f}, "
            f"var={self._var:.2f}"
            ")"
        )

    def pmf(self, x: int) -> float:
        """
        Probability Mass Function

        Parameters
        x: a value of the random variable X

        Returns
        the probability that X will take a value exactly equal to x
        """
        if x == 0:
            return 1.0 - self._p
        else:
            return self._p

    def cdf(self, x: int) -> float:
        """
        Cumulative Distribution Function

        Parameters
        x: a value of the random variable X

        Returns
        the probability that X will take a value less than or equal to x
        """
        if x == 0:
            return 1.0 - self._p
        else:
            return 1.0



from pdlpy.distribution import Distribution
from pdlpy.math import e, factorial, floor


class Poisson(Distribution):
    """
    Discrete probability distribution that expresses the probability of a given number of events occurring in a fixed interval of time or space
    """

    def __init__(self, rate: float):
        """
        Parameters
        rate: the average number of events
        """
        self._rate = rate
        self._mean = rate
        self._median = floor(rate + 1 / 3 - 0.02 / rate)
        self._mode = floor(rate)
        self._var = rate

    def __str__(self):
        return (
            "Poisson("
            f"rate={self._rate:.2f}, "
            f"mean={self._mean:.2f}, "
            f"median={self._median:.2f}, "
            f"mode={self._mode:.2f}, "
            f"var={self._var:.2f}"
            ")"
        )

    def pmf(self, x: int) -> float:
        """
        Probability Mass Function

        Parameters
        x: a value of the random variable X

        Returns
        the probability that X will take a value exactly equal to x
        """
        return (self._rate ** x) * (e ** (-self._rate)) / factorial(x)

    def cdf(self, x: int) -> float:
        """
        Cumulative Distribution Function

        Parameters
        x: a value of the random variable X

        Returns
        the probability that X will take a value less than or equal to x
        """
        if x == 0:
            return self.pmf(0)
        else:
            return self.pmf(x) + self.cdf(x - 1)
